import unittest

from i2c_registers.i2c_register_section import I2cRegisterSection


class I2cRegisterSectionTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.section = I2cRegisterSection("test", 0, 7, [0] * 8)

    def test_section_properties(self):
        self.assertEqual("TEST", self.section.name)  # add assertion here
        self.assertEqual(0, self.section.lsb)
        self.assertEqual(7, self.section.msb)
        self.assertEqual([0, 0, 0, 0, 0, 0, 0, 0], self.section.bits)

    def test_lsb_higher_than_msb(self):
        with self.assertRaises(ValueError):
            seg = I2cRegisterSection("test", 7, 0, [0] * 8)

    def test_provided_bits_too_small(self):
        with self.assertRaises(IndexError):
            seg = I2cRegisterSection("test", 0, 7, [0])

    def test_set_bits(self):
        new_bits = [1, 0, 1, 0, 1, 0, 1, 0]
        self.section.set_bits(new_bits)

        self.assertEqual([0, 1, 0, 1, 0, 1, 0, 1], self.section.bits)

    def test_set_bits_error(self):
        new_bits = [1, 0, 1, 0, 2, 0, 1, 0]
        with self.assertRaises(ValueError):
            self.section.set_bits(new_bits)


class I2cRegisterSectionConversionTestCase(unittest.TestCase):

    def test_bits_to_int(self):
        self.section = I2cRegisterSection("test", 0, 7, [0, 0, 0, 1, 0, 0, 0, 0])
        result = self.section.bits_to_int()
        self.assertEqual(16, result)

    def test_bits_to_twos_comp_int(self):
        self.section = I2cRegisterSection("test", 0, 7, [0, 0, 0, 1, 0, 0, 0, 0])
        result = self.section.bits_to_twos_comp_int()
        self.assertEqual(16, result)

    def test_bits_to_int_nibble(self):
        self.section = I2cRegisterSection("test", 0, 3, [0, 1, 0, 0])
        result = self.section.bits_to_int()
        self.assertEqual(4, result)

    def test_bits_to_twos_comp_int_nibble(self):
        self.section = I2cRegisterSection("test", 0, 3, [0, 1, 0, 0])
        result = self.section.bits_to_twos_comp_int()
        self.assertEqual(4, result)


class I2cRegisterSectionToBitsTestCase(unittest.TestCase):
    def test_i2c_register_section_to_bits(self):
        self.assertEqual([0, 0, 0, 0, 1, 1, 0, 0], I2cRegisterSection.to_bits(12, 8))

    def test_i2c_register_section_to_bits_zero(self):
        self.assertEqual([0, 0, 0, 0, 0, 0, 0, 0], I2cRegisterSection.to_bits(0, 8))


class I2cRegisterSectionToTwosCompIntTestCase(unittest.TestCase):
    def test_i2c_register_section_to_twos_comp_int(self):
        self.assertEqual(12, I2cRegisterSection.to_twos_comp_int([0, 0, 1, 1, 0, 0, 0, 0]))

    def test_i2c_register_section_to_twos_comp_int_negative(self):
        self.assertEqual(-6, I2cRegisterSection.to_twos_comp_int([0, 1, 0, 1, 1, 1, 1, 1]))


class I2cRegisterSectionNumBytesForBitsTestCase(unittest.TestCase):
    def test_lower_range(self):
        self.assertEqual(I2cRegisterSection.num_bytes_for_bits(2), 1)

    def test_upper_range(self):
        self.assertEqual(I2cRegisterSection.num_bytes_for_bits(7), 1)

    def test_simple_overflow(self):
        self.assertEqual(I2cRegisterSection.num_bytes_for_bits(9), 2)

    def test_zero(self):
        self.assertEqual(I2cRegisterSection.num_bytes_for_bits(0), 0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

from i2c_registers.i2c_register_section import I2cRegisterSection
from i2c_registers.i2c_register import I2cRegister, RegisterOperations


class I2cDevice:

    def __init__(self,
                 dev_addr: int,
                 byteorder: str,
                 i2c,
                 registers: dict[str, I2cRegister]):

        self.dev_addr = dev_addr
        self.byte_order = byteorder
        self.i2c = i2c
        self.registers = {k.upper(): v for k, v in registers.items()}

        return

    def __str__(self):
        """
        Gets a string representation of the device.
        :return: A formatted string.
        """
        out = "I2cDevice<device_address={}, registers={{\n".format(self.dev_addr)

        for k in self.registers:
            v = self.registers[k]
            # Indent output from Register.str
            v = str(v)
            v = v.split("\n")

            new_v = ""
            for i in range(0, len(v)):
                # Don't indent first line
                if i != 0:
                    new_v += "    "

                new_v += v[i]

                # No newline on last line
                if i != len(v) - 1:
                    new_v += "\n"

            out += "    {}={}\n".format(k, new_v)

        out += "}>"
        return out

    def add(self, name: str, address: int, op_mode: RegisterOperations, signed: bool,
            sections: dict[str, I2cRegisterSection]):
        """
        Adds a register to the device.
        :param name: The name of the register.
        :param address: The address of the register
        :param op_mode: The operation mode: Read, Write, ReadWrite
        :param signed: An indicator to determine if the register value iss signed.
        :param sections: The sections of the Register. (Optional; can pass in {})
        :return: A Register object with specified name.
        """
        key = name.upper()
        if key in self.registers:
            raise KeyError("I2cRegister with name already exists. name: {}".format(name))

        register = I2cRegister(name.upper(), address, op_mode, signed, sections)
        register.set_parent(self)

        self.registers[key] = register

        return self.registers[key]

    def get(self, name: str, read_first: bool = False) -> I2cRegister:
        """
        Gets a Register by name.
        :param name: The register name.
        :param read_first: Optional. Perform a physical read first.
        :return: The I2cRegister for the name provided.
        :raises KeyError: If the section does not exist.
        """
        key = name.upper()
        if key not in self.registers:
            raise KeyError("Register with name \"{}\" not found".format(name))

        # Read first if asked
        if read_first:
            self.read(key)

        return self.registers[key]

    def read(self, name: str):
        """
        Performs a physical read for the register name provided.

        :param name: The register name
        """
        self.get(name).read()

        return

    def write(self, name: str):
        """
        Performs a physical write for the register name provided.

        :param name: The register name
        """
        self.get(name).write()

        return



import unittest

from i2c_registers.i2c_register_section import I2cRegisterSection
from i2c_registers.i2c_register import I2cRegister, RegisterOperations
from i2c_registers.i2c_device import I2cDevice


class I2cDeviceTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.section = I2cRegisterSection("test_seg", 0, 7, [0] * 8)
        self.register = I2cRegister("test_reg", 0x03,
                                    RegisterOperations.ReadWrite, False, {"test": self.section})

        self.device = I2cDevice(0x62, "big", None, {"test_reg": self.register})

    def test_properties(self):
        self.assertEqual(0x62, self.device.dev_addr)
        self.assertEqual("big", self.device.byte_order)
        self.assertEqual(None, self.device.i2c)

    def test_get(self):
        register = self.device.get("test_reg")
        self.assertEqual(0x03, register.reg_addr)
        return


if __name__ == '__main__':
    unittest.main()



import unittest

from i2c_registers.i2c_register import I2cRegister, RegisterOperations
from i2c_registers.i2c_register_section import I2cRegisterSection


class I2cRegisterTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.section = I2cRegisterSection("test", 0, 7, [0] * 8)
        self.register = I2cRegister("test", 0x03,
                                    RegisterOperations.ReadWrite, False, {"test": self.section})

    def test_properties(self):
        self.assertEqual("TEST", self.register.name)
        self.assertEqual(3, self.register.reg_addr)
        self.assertEqual(RegisterOperations.ReadWrite, self.register.op_mode)
        self.assertEqual({"TEST": self.section}, self.register.sections)

    def test_get_section(self):
        seg = self.register.get_section("test")
        self.assertEqual("TEST", seg.name)
        self.assertEqual(0, seg.lsb)
        self.assertEqual(7, seg.msb)
        bits = seg.get_bits()
        self.assertEqual([0, 0, 0, 0, 0, 0, 0, 0], bits)

    def test_not_exist(self):
        with self.assertRaises(KeyError):
            self.register.get_section("DOES_NOT_EXIST")

    def test_len_bytes(self):
        result = self.register.len_bytes()
        self.assertEqual(1, result)

    def test_set_value_8bit(self):
        self.register.set_value(16)
        seg = self.register.get_section("test")
        bits = seg.get_bits()
        self.assertEqual([0, 0, 0, 1, 0, 0, 0, 0], bits)


class I2cRegisterSetGetValueTestCase(unittest.TestCase):

    def test_set_value_8bit(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 7, [0, 0, 0, 0, 0, 0, 0, 0])
        register.set_value(16)
        seg = register.get_section("test_seg")
        bits = seg.get_bits()
        self.assertEqual([0, 0, 0, 1, 0, 0, 0, 0], bits)

    def test_set_value_16bit(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 15,
                             [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        register.set_value(31135)
        seg = register.get_section("test_seg")

        # 799F
        # 31135
        # 0111 1001 1001 1111
        self.assertEqual([1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 0], seg.bits)  # Actual bits are reversed

        bits = seg.get_bits()
        self.assertEqual([0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1], bits)

    def test_get_value_8bit(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 7, [0, 0, 0, 1, 0, 0, 0, 0])

        reg_val = register.get_value()
        self.assertEqual(16, reg_val)

    def test_get_value_16bit(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 15,
                             [0, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1])
        # 799F
        # 31135
        # 0111 1001 1001 1111
        reg_val = register.get_value()
        self.assertEqual(31135, reg_val)

    def test_get_value_all_bits_8bit(self):
        register = I2cRegister("IODIR", 0x00, RegisterOperations.ReadWrite, False, {})
        register.add_section("IO7", 7, 7, [0])
        register.add_section("IO6", 6, 6, [0])
        register.add_section("IO5", 5, 5, [0])
        register.add_section("IO4", 4, 4, [1])
        register.add_section("IO3", 3, 3, [0])
        register.add_section("IO2", 2, 2, [0])
        register.add_section("IO1", 1, 1, [0])
        register.add_section("IO0", 0, 0, [0])

        reg_val = register.get_value()
        self.assertEqual(16, reg_val)

    def test_get_value_all_bits_16bit(self):
        register = I2cRegister("CONF", 0x00, RegisterOperations.ReadWrite, False, {})
        register.add_section("RST", 15, 15, [0])
        register.add_section("BRNG", 13, 14, [1, 1])
        register.add_section("PG", 11, 12, [1, 1])
        register.add_section("BADC", 7, 10, [0, 0, 1, 1])
        register.add_section("SADC", 3, 6, [0, 0, 1, 1])
        register.add_section("MODE", 0, 2, [1, 1, 1])

        # 799F
        # 31135
        # 0111 1001 1001 1111 -> 1111 1001 1001 1110

        reg_val = register.get_value()
        self.assertEqual(31135, reg_val)

    def test_set_value_8bit_get_value(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 7, [0, 0, 0, 0, 0, 0, 0, 0])
        register.set_value(16)
        seg = register.get_section("test_seg")
        self.assertEqual([0, 0, 0, 0, 1, 0, 0, 0], seg.bits)  # Actual bits are reversed
        bits = seg.get_bits()
        self.assertEqual([0, 0, 0, 1, 0, 0, 0, 0], bits)

        value = register.get_value()
        self.assertEqual(16, value)

    def test_get_bit_13(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 1, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 0, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])
        register.get_value()
        bit_value = register.get_bit(13)
        self.assertEqual(1, bit_value)

    def test_get_bit_5(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 0, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 1, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])
        register.get_value()
        bit_value = register.get_bit(5)
        self.assertEqual(1, bit_value)

    def test_clear_bit_5(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 0, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 1, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])

        register.clear_bit(5)
        reg_val = register.get_value()
        self.assertEqual(0, reg_val)

    def test_set_section_bits(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 0, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 0, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])

        register.set_section_bits("Sec1", [1, 0, 0, 0])
        reg_val = register.get_value()
        self.assertEqual(128, reg_val)

    def test_set_bit_5(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 0, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 0, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])

        register.set_bit(5)
        reg_val = register.get_value()
        self.assertEqual(32, reg_val)

    def test_set_bit_10(self):
        register = I2cRegister("REG", 0x00, RegisterOperations.ReadWrite, False, {})

        register.add_section("Sec3", 12, 15, [0, 0, 0, 0])
        register.add_section("Sec2", 8, 11, [0, 0, 0, 0])
        register.add_section("Sec1", 4, 7, [0, 0, 0, 0])
        register.add_section("Sec0", 0, 3, [0, 0, 0, 0])

        register.set_bit(10)
        reg_val = register.get_value()
        self.assertEqual(1024, reg_val)

    def test_get_value_duplicate_bits(self):
        register = I2cRegister("IODIR", 0x00, RegisterOperations.ReadWrite, False, {})
        register.add_section("IODIR", 0, 7, [0] * 8)
        register.add_section("IO7", 7, 7, [0])
        register.add_section("IO6", 6, 6, [0])
        register.add_section("IO5", 5, 5, [0])
        register.add_section("IO4", 4, 4, [0])
        register.add_section("IO3", 3, 3, [0])
        register.add_section("IO2", 2, 2, [0])
        register.add_section("IO1", 1, 1, [0])
        register.add_section("IO0", 0, 0, [0])

        with self.assertRaises(KeyError):
            register.get_value()

    def test_get_value_missing_bits(self):
        register = I2cRegister("IOCON", 0x00, RegisterOperations.ReadWrite, False, {})
        register.add_section("SREAD", 5, 5, [0])
        register.add_section("DISSLW", 4, 4, [0])
        register.add_section("ODR", 2, 2, [0])
        register.add_section("INTPOL", 1, 1, [0])

        with self.assertRaises(SyntaxError):
            register.get_value()

    def test_get_value_not_used_bits(self):
        register = I2cRegister("IOCON", 0x00, RegisterOperations.ReadWrite, False, {})
        register.add_not_used_section(6, 7)
        register.add_section("SREAD", 5, 5, [0])
        register.add_section("DISSLW", 4, 4, [1])
        register.add_not_used_section(3, 3)
        register.add_section("ODR", 2, 2, [0])
        register.add_section("INTPOL", 1, 1, [0])
        register.add_not_used_section(0, 0)

        reg_val = register.get_value()
        self.assertEqual(16, reg_val)


class I2cRegisterAddTestCase(unittest.TestCase):

    def test_perfect(self):
        register = I2cRegister("test", 0x03, RegisterOperations.ReadWrite, False, {})
        register.add_section("test_seg", 0, 2, [0] * 3)

        section = register.get_section("test_seg")
        self.assertEqual("TEST_SEG", section.name)
        self.assertEqual(0, section.lsb)
        self.assertEqual(2, section.msb)
        self.assertEqual([0, 0, 0], section.bits)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
from enum import Flag, auto

from i2c_registers.i2c_register_section import I2cRegisterSection


class RegisterOperations(Flag):
    Read = auto()
    Write = auto()
    ReadWrite = Read | Write


class I2cRegister:
    __parent = None
    __managing_sections = {}

    def __init__(self,
                 name: str,
                 reg_addr: int,
                 op_mode: RegisterOperations,
                 signed: bool,
                 sections: dict[str, I2cRegisterSection]):
        self.name = name.upper()
        self.reg_addr = reg_addr
        self.op_mode = op_mode
        self.signed = signed
        self.sections = {k.upper(): v for k, v in sections.items()}

        return

    @property
    def value(self) -> int:
        return self.get_value()

    def __len__(self):
        """
        Gets the length of register in bits.
        :return: The total number of bits for the register.
        """
        result = 0
        for section in self.sections:
            result += len(self.sections[section])

        return result

    def __str__(self):
        """
        Gets a string representation of the register.
        :return: A formatted string.
        """
        result = "Register<name={}, address={}, op_mode={}, sections={{\n"
        result.format(self.name, self.reg_addr, self.op_mode.name)
        for key in self.sections:
            section = self.sections[key]
            result += "    {}={}\n".format(key, str(section))

        result += "}>"

        return result

    def set_parent(self, parent):
        """
        Sets the Parent Device for the register.
        :param parent: The parent Device (I2cDevice)
        """
        self.__parent = parent

        return

    def len_bytes(self):
        """
        Length of the register in bytes.
        :return: The total number of bytes for the register.
        """
        return I2cRegisterSection.num_bytes_for_bits(len(self))

    def get_section(self, name: str) -> I2cRegisterSection:
        """
        Gets a Register Section by name.
        :param name: The section name.
        :return: The I2cRegisterSection for the name provided.
        :raises KeyError: If the section does not exist.
        """
        key = name.upper()
        if key not in self.sections:
            raise KeyError("No section found with name: \"{}\"".format(name))

        self._check_sections()
        return self.sections[key]

    def add_section(self, name: str, lsb: int, msb: int, bits: list[int]):
        """
        Add Register Section

        :param name: Name of the section
        :param lsb: Index of LSB (the least significant bit)
        :param msb: Index of MSB (the most significant bit)
        :param bits: List of bits, each element of list is either 0 or 1. The bit order: 7 6 5 4 3 2 1 0
        :return: The current Register, for chain adding of I2cRegisterSection to same Register
        :raises KeyError: If the sections has already been added
        """
        key = name.upper()
        if key in self.sections:
            raise KeyError("I2cRegisterSection with name already exists. name: {}".format(name))

        self.sections[key] = I2cRegisterSection(name.upper(), lsb, msb, bits)

        return self

    def add_not_used_section(self, lsb: int, msb: int):
        """
        Add Not Used Register Section

        :param lsb: Index of LSB (the least significant bit)
        :param msb: Index of MSB (the most significant bit)
        :return: The current Register, for chain adding of I2cRegisterSection to same Register
        """
        name = "NU{}_{}".format(lsb, msb)
        self.add_section(name, lsb, msb, [0] * (msb - lsb + 1))

        return self

    def set_section_bits(self, name: str, bits: list[int]) -> None:
        """
        Sets all the bits for the section
        :param name: Name of the section
        :param bits: List of bits, each element of list is either 0 or 1. The bit order: 7 6 5 4 3 2 1 0
        :return: None
        """
        self.get_section(name).set_bits(bits)

        return

    def set_bit(self, bit_index: int) -> None:
        """
        Sets the bit to 1 for the given bit index
        :param bit_index: The bit index of the bit to set
        :return: None
        """
        section = self.get_section_for_index(bit_index)
        section_bit_index = section.get_index(bit_index)
        section.bits[section_bit_index] = 1
        return

    def clear_bit(self, bit_index: int) -> None:
        """
        Clears the bit to 0 for the given bit index
        :param bit_index: The bit index of the bit to set
        :return: None
        """
        section = self.get_section_for_index(bit_index)
        section_bit_index = section.get_index(bit_index)
        section.bits[section_bit_index] = 0
        return

    def get_bit(self, bit_index: int) -> int:
        """
        Gets the value of the bit for the register bit index.
        :param bit_index: The register bit index to retrieve.
        :return: The value for the bit index.
        """
        section = self.get_section_for_index(bit_index)
        section_bit_index = section.get_index(bit_index)
        return section.bits[section_bit_index]

    def set_value(self, value: int, write_after: bool = False) -> None:
        """
        Sets the value of the register.
        :param value: The value to set the register.
        :param write_after: Optional. Write to the physical device after setting the value.
        """
        bytes_count = self.len_bytes()

        value_bytes = value.to_bytes(bytes_count, "big", signed=self.signed)
        value_bytes = value_bytes[::-1]

        for section in self.sections:
            self.sections[section].update_bits(value_bytes)

        if write_after:
            self.write()

        return

    def get_value(self, read_first: bool = False) -> int:
        """
        Gets the value of the register.
        :param read_first: Optional. Perform a physical read first.
        :return: The value of the register.
        """
        if read_first:
            self.read()

        bytes_arr = self._get_bytes()
        reg_val = bytearray(bytes_arr)

        result = int.from_bytes(reg_val, byteorder="little", signed=self.signed)

        return result

    def read(self):
        """
        Performs a physical read of the register.
        :raises SystemError: If the parent device has not been set.
        :raises SystemError: If the read from the physical device fails.
        :raises AttributeError: If the register is marked as write only.
        """
        if RegisterOperations.Read & self.op_mode == RegisterOperations.Read:
            # Get number of bytes to read, will raise AssertionError if sections do not create round number of bytes
            bytes_count = self.len_bytes()
            reg_bytes: list[int]
            if self.__parent is None:
                raise SystemError("No I2C Device set as parent")

            try:
                reg_bytes = self.__parent.i2c.read_bytes(self.__parent.dev_addr, self.reg_addr, bytes_count)

                if self.__parent is not None:
                    if bytes_count > 1 and self.__parent.byte_order == "big":
                        reg_bytes.reverse()

            except Exception as e:
                raise SystemError("Failed to read i2c: {}".format(e))

            # Loop through each byte read and map to elements in RegisterSection bit arrays
            for section in self.sections:
                self.sections[section].update_bits(reg_bytes)  # Raises KeyError if we didn't read enough bytes
        else:
            msg = "Register {} is not set up to allow read operations, op_mode: \"{}\""
            raise AttributeError(msg.format(self.name, self.op_mode.name))

        return

    def write(self):
        """
        Performs a physical read of the register.

        :raises SystemError: If the parent device has not been set.
        :raises SystemError: If the read from the physical device fails.
        :raises AttributeError: If the register is marked as write only.
        """
        if RegisterOperations.Write & self.op_mode == RegisterOperations.Write:

            if self.__parent is None:
                raise SystemError("No I2C Device set as parent")

            reg_bytes = self._get_bytes()
            try:
                bytes_count = len(reg_bytes)
                if self.__parent is not None:
                    if bytes_count > 1 and self.__parent.byte_order == "big":
                        reg_bytes.reverse()

                # Write to i2c
                write_status = self.__parent.i2c.write_bytes(self.__parent.dev_addr, self.reg_addr, reg_bytes)
            except Exception as e:
                raise SystemError("Failed to write to i2c: {}".format(e))

            if write_status == 1:
                raise SystemError("Failed to write to i2c register: {}".format(self.name))

            return
        else:
            msg = "Register {} is not set up to allow write operations, op_mode: \"{}\""
            raise AttributeError(msg.format(self.name, self.op_mode.name))

    def get_section_for_index(self, bit_index: int) -> I2cRegisterSection:
        """
        Gets a Register Section for the bit index.
        :param bit_index: The register bit index
        :return: The I2cRegisterSection for the bit index provided.
        """
        self._check_sections()
        sec_key = self.__managing_sections[bit_index][0]
        return self.sections.get(sec_key)

    def _check_sections(self):

        bits = {}
        max_bit = 0
        managing_sections = {}

        for section in self.sections:
            section = self.sections[section]
            for bit_ix in range(len(section.bits)):
                actual_bit_ix = bit_ix + section.lsb
                # Record maximum bit index for continuous test later
                if actual_bit_ix > max_bit:
                    max_bit = actual_bit_ix

                # Mark which section manages the bit in case of configuration error
                if actual_bit_ix not in managing_sections:
                    managing_sections[actual_bit_ix] = []
                managing_sections[actual_bit_ix].append(section.name)

                bits[actual_bit_ix] = section.bits[bit_ix]

        # Check bits are continuous and only 1 section controls each bit
        cont_check_err_is = []  # Indexes where bit array is not continuous
        managing_section_check_err_is = []  # Indexes where more than 1 RegisterSection is controlling a bit
        for bit_ix in range(max_bit + 1):
            # Continuous check
            if bit_ix not in bits:
                cont_check_err_is.append(bit_ix)

            # Managing Section check
            if (bit_ix in managing_sections) and (len(managing_sections[bit_ix]) > 1):
                managing_section_check_err_is.append(bit_ix)

        # Raise errors if there are any
        if len(cont_check_err_is) > 0:
            msg = "I2cRegisterSection are not configured to make a continuous series of bits, \
                        no values at indexes: {}"
            raise SyntaxError(msg.format(cont_check_err_is))

        if len(managing_section_check_err_is) > 0:
            indexes_msg = ""

            for i in range(len(managing_section_check_err_is)):
                bit_ix = managing_section_check_err_is[i]
                indexes_msg += "{} (competing sections: {})".format(bit_ix, managing_sections[bit_ix])

                if i != len(managing_section_check_err_is) - 1:
                    indexes_msg += ", "

            msg = "More than one I2cRegisterSection is managing the following bit indexes: {}"
            raise KeyError(msg.format(indexes_msg))

        if not self.__managing_sections:
            self.__managing_sections = managing_sections

        return bits, max_bit

    def _get_bytes(self) -> list[int]:

        bits, max_bit = self._check_sections()

        # Convert from bits map to bits array
        bits_arr = [bits[key] for key in sorted(bits.keys())]

        # Create bytes array
        bytes_arr = I2cRegisterSection.to_padded_byte_arr(bits_arr)

        return bytes_arr



#!/usr/bin/python3
import math


class I2cRegisterSection:

    @staticmethod
    def to_int(bits):
        result = 0
        for bit in bits:
            result = (result << 1) | bit

        return result

    @staticmethod
    def to_bits(number, size):
        bits = size
        if number > 0:
            bits = int(max(size, int(math.log(number, 2)) + 1))

        result = [1 if number & (1 << (bits - 1 - n)) else 0 for n in range(bits)]

        return result

    @staticmethod
    def to_twos_comp_int(bits):
        bits_str = ""

        for b in reversed(bits):
            bits_str += str(b)

        size = len(bits)
        result = int(bits_str, 2)

        if (result & (1 << (size - 1))) != 0:
            result = result - (1 << size)

        return result

    @staticmethod
    def num_bytes_for_bits(bits):
        return int(math.ceil(float(bits) / 8.0))

    @staticmethod
    def to_padded_byte_arr(bits):
        result = []
        byte_slice_lower = 0  # Increases by 8 for each byte

        # Determine how many bytes the provided bits are and loop that many times
        # Each loop "synthesizes" a new byte from the bits array
        for byte_i in range(I2cRegisterSection.num_bytes_for_bits(len(bits))):
            # Check that upper limit isn't too big
            byte_slice_upper = ((byte_i + 1) * 8) - 1  # The index we *want* to slice to
            to_pad = 0  # Used to keep track of how many 0s we need to pad the end of this byte
            if byte_slice_upper > len(bits) - 1:
                # Keep track of the fact that we need to pad the end of this byte with some 0s
                to_pad = byte_slice_upper - (len(bits) - 1)

                # Resize if index we wanted to slice to is too big
                byte_slice_upper = len(bits) - 1

            # Convert
            # Add 1 to byte_slice_upper because upper range of slice is not inclusive
            byte_slice = bits[byte_slice_lower:byte_slice_upper + 1]

            # Append padding
            if to_pad > 0:
                byte_slice.extend([0] * to_pad)

            byte = I2cRegisterSection.to_int(byte_slice[::-1])
            result.append(byte)

            # Add 8 to lower slice bound for next byte
            byte_slice_lower += 8

        return result

    def __init__(self,
                 name: str,
                 lsb: int,
                 msb: int,
                 bits: list[int]) -> None:
        """
        Creates a Register Section instance

        :param name: Name of section
        :param lsb: Index of LSB (the least significant bit)
        :param msb: Index of MSB (the most significant bit)
        :param bits: List of bits, each element of list is either 0 or 1. The bit order: 7 6 5 4 3 2 1 0
        :raises IndexError: If length of bits array is less than that defined by lsb and msb
        :raises ValueError: If lsb or msb is not in the range [0, 7] or lsb and msb are greater or
                          less than each other in wrong way
        """

        self.name = name.upper()

        # Sanity check LSB and MSB indexes
        if lsb > msb:
            msg = "LSB index can not be greater than MSB index, lsb_i: {}, msb_i: {}"
            raise ValueError(msg.format(lsb, msb))

        self.lsb = lsb
        self.msb = msb
        self.bits: list[int] = []
        _bits = bits[:]
        self.set_bits(_bits)

    def __len__(self):
        """
        Get length of register section.

        Uses lsb and msb to calculate.
        :return: The Number of bits represented by RegisterSection.
        """
        return self.msb - self.lsb + 1

    def __str__(self):
        """
        Gets a string representation of the section.
        :return: A formatted string.
        """
        msg = "RegisterSegment<name={}, lsb={}, msb={}, bits={}>"
        bits = self.get_bits()
        return msg.format(self.name, self.lsb, self.msb, bits)

    def bits_to_int(self):
        _bits = self.bits[::-1]
        return I2cRegisterSection.to_int(_bits)

    def bits_to_twos_comp_int(self):
        return I2cRegisterSection.to_twos_comp_int(self.bits)

    def update_bits(self, reg_bytes) -> None:
        """
        Updates the bits for the section from byte values
        :param reg_bytes: The register byte values e.g. [[0xFE],[0xFE]]
        """
        # Check that bytes array contains values inside lsb and msb range
        min_bytes = I2cRegisterSection.num_bytes_for_bits(self.msb + 1)
        if len(reg_bytes) < min_bytes:
            msg = "Provided bytes array does not contain enough bytes to fill MSB, bytes: {}, \
            MSB index: {}, required bytes length: {}"

            raise KeyError(msg.format(reg_bytes, self.msb, min_bytes))

        # Determine start and end byte by dividing lsb & msb by 8 and rounding down
        start_byte = int(math.floor(float(self.lsb) / 8.0))
        end_byte = int(math.floor(float(self.msb) / 8.0))

        # Convert needed bytes into bits
        # Keys will be offset by start_byte
        needed_bytes_as_bits = []

        for byte_ix in range(start_byte, end_byte + 1):
            byte = reg_bytes[byte_ix]
            converted_bits = I2cRegisterSection.to_bits(byte, 8)

            # Convert bits and check bits are all 0 or 1
            for i in range(len(converted_bits)):
                converted_bits[i] = int(converted_bits[i])

            needed_bytes_as_bits.append(converted_bits[::-1])

        # Loop through bits
        for bit_ix in range(self.lsb, self.msb + 1):
            in_byte_i = int(math.floor(float(bit_ix) / 8.0))
            bit_offset = (in_byte_i * 8)  # Used to figure out which bit in the byte we are in

            self.bits[bit_ix - self.lsb] = needed_bytes_as_bits[in_byte_i - start_byte][bit_ix - bit_offset]

        return

    def set_bits(self, bits: list[int]):
        """
        Set section bits

        Runs some sanity checks on the new bits before setting them.

        :param bits: List of bits, each element of list is either 0 or 1. The bit order: 7 6 5 4 3 2 1 0
        :raises IndexError: If length of bits array is less than that defined by lsb and msb
        :raises ValueError: If an element of the provided bits array is not equal to 0 or 1
        """
        if len(bits) != len(self):
            msg = "Default list must be size that specified by lsb and msb, was: {}, should be: {}"
            raise IndexError(msg.format(len(bits), len(self)))

        is_valid_list = [True if b == 1 or b == 0 else False for b in bits]

        if False in is_valid_list:
            msg = "Bits can only have the integer values 0 or 1"
            raise ValueError(msg)

        _bits = bits[:]  # Copy List -> bit order => 7 6 5 4 3 2 1 0
        rev_bits = _bits[::-1]  # Reverse List -> bit order => 0 1 2 3 4 5 6 7
        self.bits = rev_bits  # Assign to property/ attribute

    def get_bits(self) -> list[int]:
        """
        Gets the section bits.
        :return: List of bits, each element of list is either 0 or 1. The bit order: 7 6 5 4 3 2 1 0
        """
        _bits = self.bits[:]

        return _bits[::-1]

    def get_index(self, bit_index):
        """
        Gets the section index from the register bit index
        :param bit_index: The register bit index. The bit order: 7 6 5 4 3 2 1 0
        :return: The corresponding section bit index
        """
        return bit_index - self.lsb
